import click
import os
from dialog_lib.agents import DialogOpenAI, DialogAnthropic
from dialog_lib.memory import generate_local_memory_instance

@click.group()
def cli():
    pass


def run_llm(agent, instance_name, memory, debug=False):
    click.echo("## Welcome to talkd.ai/dialog-lib CLI")
    click.echo("## Initialized the LLM process")

    while True:
        input_text = input("You: ")

        if input_text == "q":
            break

        output_text = agent.process(input_text)
        print(f"{instance_name}: {output_text}")

    if debug:
        click.echo(memory.chat_memory.messages)



@cli.command()
@click.option("--model", default="gpt-3.5-turbo", help="The model to use")
@click.option("--temperature", default=0.1, help="The temperature for generating responses")
@click.option("--llm-api-key", default=os.environ.get("OPENAI_API_KEY"), help="The OpenAI API key")
@click.option("--prompt", default="You are a bot called Sara. Be nice to other human beings.", help="The prompt for the dialog")
@click.option("--debug", default=False, help="Prints the memory after the dialog ends", is_flag=True)
def openai(model, temperature, llm_api_key, prompt, debug):
    memory = generate_local_memory_instance()
    dialog = DialogOpenAI(
        model=model,
        temperature=temperature,
        llm_api_key=llm_api_key,
        prompt=prompt,
        memory=memory,
    )
    run_llm(dialog, "ChatGPT", memory, debug=debug)


@cli.command()
@click.option("--model", default="claude-3-opus-20240229", help="The model to use")
@click.option("--temperature", default=0, help="The temperature for generating responses")
@click.option("--llm-api-key", default=os.environ.get("ANTHROPIC_API_KEY", None), help="The Anthropic API key", required=True)
@click.option("--prompt", default="You are a bot called Sara. Be nice to other human beings.", help="The prompt for the dialog")
@click.option("--debug", default=False, help="Prints the memory after the dialog ends", is_flag=True)
def anthropic(model, temperature, llm_api_key, prompt, debug):
    memory = generate_local_memory_instance()
    dialog = DialogAnthropic(
        model=model,
        temperature=temperature,
        llm_api_key=llm_api_key,
        prompt=prompt,
        memory=memory,
    )
    run_llm(dialog, "Anthropic", memory, debug=debug)


def main():
    cli()



import os
import pytest
import sqlalchemy

from aioresponses import aioresponses
from sqlalchemy.orm import Session
from dialog_lib.db.models import Base


@pytest.fixture
def db_engine():
    return sqlalchemy.create_engine(os.environ.get('DATABASE_URL'))

@pytest.fixture
def db_session(db_engine):
    Base.metadata.create_all(db_engine)
    session = Session(db_engine)
    yield session
    session.rollback()
    session.close()

@pytest.fixture
def mock_aioresponse():
    with aioresponses() as m:
        yield m


from dialog_lib.db.memory import *
from langchain.memory.buffer import ConversationBufferMemory

def generate_local_memory_instance(memory_key="chat_history"):
    return ConversationBufferMemory(
        memory_key=memory_key,
        return_messages=True
    )




from typing import List

from sqlalchemy import select
from langchain_core.embeddings import Embeddings
from dialog_lib.db.models import CompanyContent


def generate_embeddings(
    documents: List[str], embedding_llm_instance: Embeddings = None
):
    """
    Generate embeddings for a list of documents
    """
    return embedding_llm_instance.embed_documents(documents)


def generate_embedding(document: str, embedding_llm_instance: Embeddings = None):
    """
    Generate embeddings for a single instance of document

    :param document: str - The document/string to generate embeddings for
    :param embedding_llm_instance: - The Embedding LLM instance to use for generating embeddings

    """
    return embedding_llm_instance.embed_query(document)


def get_most_relevant_contents_from_message(
    message,
    top=5,
    dataset=None,
    session=None,
    embeddings_llm=None,
    cosine_similarity_threshold=0.5,
    model=CompanyContent,
    embedding_column="embedding",
):
    message_embedding = generate_embedding(message, embeddings_llm)
    filters = [
        model.embedding.cosine_distance(message_embedding)
        < cosine_similarity_threshold,
    ]

    if dataset is not None:
        filters.append(model.dataset == dataset)

    possible_contents = session.scalars(
        select(model)
        .filter(*filters)
        .order_by(getattr(model, embedding_column).cosine_distance(message_embedding))
        .limit(top)
    ).all()
    return possible_contents



from dialog_lib.tests.fixtures.embedding_model import FakeEmbeddingModel
from dialog_lib.embeddings.generate import generate_embedding, generate_embeddings

def test_generate_embedding_for_single_document():
    embedding_model = FakeEmbeddingModel()
    embedding = generate_embedding("Hello, world!", embedding_model)
    assert len(embedding) == 1536

def test_generate_embedding_for_multiple_documents():
    embedding_model = FakeEmbeddingModel()
    embeddings = generate_embeddings(["Hello, world!", "Hello, world 2!"], embedding_model)
    assert len(embeddings) == 2
    assert len(embeddings[0]) == 1536


import pytest
import responses

from dialog_lib.db.models import CompanyContent
from dialog_lib.loaders.web import load_webpage


def test_load_web_content(mock_aioresponse, db_session, mocker):
    mocker.patch('dialog_lib.loaders.web.generate_embedding', return_value=[0] * 1536)
    mock_aioresponse.get('http://example.com', body='Hello, world!')

    content = load_webpage('http://example.com', db_session, None, 1)
    assert content.question == "Example Domain"
    assert content.embedding == [0] * 1536

    content = db_session.query(CompanyContent).all()
    assert len(content) == 1



import os
import pytest
from dialog_lib.agents.openai import DialogOpenAI
from dialog_lib.agents.anthropic import DialogAnthropic

def test_openai_agent():
    os.environ['OPENAI_API_KEY'] = 'sk_test_1234567890'
    agent = DialogOpenAI()
    assert agent is not None

def test_anthropic_agent():
    agent = DialogAnthropic()
    assert agent is not None


import uuid

from sqlalchemy import Table, MetaData
from sqlalchemy import Column, Integer, DateTime, String, text, Text
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.dialects.postgresql import UUID, JSONB

from pgvector.sqlalchemy import Vector


class Base(DeclarativeBase):
    pass


class ChatMessages(Base):
    __tablename__ = "chat_messages"

    id = Column(Integer, nullable=False, primary_key=True, autoincrement=True)
    parent = Column(Integer, nullable=True)
    session_id = Column(String, nullable=False)
    message = Column(JSONB, nullable=False)
    timestamp = Column(
        DateTime, nullable=False, server_default=text("CURRENT_TIMESTAMP")
    )


class Chat(Base):
    __tablename__ = "chats"

    session_id = Column(
        String, nullable=False, default=str(uuid.uuid4()), primary_key=True
    )
    tags = Column(String, nullable=True)


class CompanyContent(Base):
    __tablename__ = "contents"

    id = Column(Integer, nullable=False, primary_key=True, autoincrement=True)
    category = Column(String, nullable=False)
    subcategory = Column(String, nullable=False)
    question = Column(String, nullable=False)
    content = Column(Text, nullable=False)
    embedding = Column(Vector(1536), nullable=False)
    dataset = Column(String, nullable=True)
    link = Column(String, nullable=True)



from .abstract import AbstractDialog
from langchain_openai.chat_models.base import ChatOpenAI


class DialogOpenAI(AbstractDialog):
    def __init__(self, *args, **kwargs):
        model = kwargs.pop("model", "gpt-3.5-turbo")
        temperature = kwargs.pop("temperature", 0.1)
        kwargs["model_class"] = ChatOpenAI(
            model=model,
            temperature=temperature,
            openai_api_key=kwargs.get("llm_api_key"),
        )
        super().__init__(*args, **kwargs)

    def postprocess(self, output):
        return output.get("text")


import psycopg

from langchain_postgres import PostgresChatMessageHistory
from langchain.schema.messages import BaseMessage, _message_to_dict

from .models import Chat, ChatMessages


class CustomPostgresChatMessageHistory(PostgresChatMessageHistory):
    """
    Custom chat message history for LLM
    """

    def __init__(
        self,
        *args,
        parent_session_id=None,
        dbsession=None,
        chats_model=Chat,
        chat_messages_model=ChatMessages,
        ssl_mode=None,
        **kwargs,
    ):
        self.parent_session_id = parent_session_id
        self.dbsession = dbsession
        self.chats_model = chats_model
        self.chat_messages_model = chat_messages_model
        self._connection = psycopg.connect(
            kwargs.pop("connection_string"), sslmode=ssl_mode
        )
        self._session_id = kwargs.pop("session_id")
        self._table_name = kwargs.pop("table_name")


    def create_tables(self) -> None:
        """
        create table if it does not exist
        add a new column for timestamp
        """
        create_table_query = f"""CREATE TABLE IF NOT EXISTS {self.table_name} (
            id SERIAL PRIMARY KEY,
            session_id TEXT NOT NULL,
            message JSONB NOT NULL,
            timestamp TIMESTAMP NOT NULL DEFAULT CURRENT_TIMESTAMP
        );"""
        self.cursor.execute(create_table_query)
        self.connection.commit()

    def add_tags(self, tags: str) -> None:
        """Add tags for a given session_id/uuid on chats table"""
        self.dbsession.query(self.chats_model).where(
            self.chats_model.session_id == self._session_id
        ).update({getattr(self.chats_model, "tags"): tags})
        self.dbsession.commit()

    def add_message(self, message: BaseMessage) -> None:
        """Append the message to the record in PostgreSQL"""
        message = self.chat_messages_model(
            session_id=self._session_id, message=_message_to_dict(message)
        )
        if self.parent_session_id:
            message.parent = self.parent_session_id
        self.dbsession.add(message)
        self.dbsession.commit()


def generate_memory_instance(
    session_id,
    parent_session_id=None,
    dbsession=None,
    database_url=None,
    chats_model=Chat,
    chat_messages_model=ChatMessages,
):
    """
    Generate a memory instance for a given session_id
    """

    return CustomPostgresChatMessageHistory(
        connection_string=database_url,
        session_id=session_id,
        parent_session_id=parent_session_id,
        table_name="chat_messages",
        dbsession=dbsession,
        chats_model=chats_model,
        chat_messages_model=chat_messages_model,
    )


def add_user_message_to_message_history(
    session_id, message, memory=None, dbsession=None, database_url=None
):
    """
    Add a user message to the message history and returns the updated
    memory instance
    """
    if not memory:
        memory = generate_memory_instance(
            session_id, dbsession=dbsession, database_url=database_url
        )

    memory.add_user_message(message)
    return memory


def get_messages(session_id, dbsession=None, database_url=None):
    """
    Get all messages for a given session_id
    """
    memory = generate_memory_instance(
        session_id, dbsession=dbsession, database_url=database_url
    )
    return memory.messages



from typing import TYPE_CHECKING, Any, Dict, List, Optional

from dialog_lib.db.models import CompanyContent
from sqlalchemy.orm import DeclarativeBase, Session

from langchain_core.retrievers import BaseRetriever
from langchain_core.documents import Document
from dialog_lib.embeddings.generate import get_most_relevant_contents_from_message

class DialogRetriever(BaseRetriever):
    content_model: DeclarativeBase = CompanyContent
    session: Session
    threshold: float = 0.5
    dataset: Optional[Any] = None
    embedding_llm: Optional[Any] = None
    embedding_column: str = "embedding"
    top_k: int = 5

    def _get_relevant_documents(self, query, *, run_manager):
        relevant_contents = get_most_relevant_contents_from_message(
            query,
            top=self.top_k,
            dataset=self.dataset,
            session=self.session,
            embeddings_llm=self.embedding_llm,
            cosine_similarity_threshold=self.threshold,
            model=self.content_model,
            embedding_column=self.embedding_column,
        )
        return [
            Document(
                page_content=f"{content.question}\n\n{content.content}",
                metadata={
                    "title": content.question,
                    "category": content.category,
                    "subcategory": content.subcategory,
                    "dataset": content.dataset,
                    "link": content.link,
                },
            )
            for content in relevant_contents
        ]





from dialog_lib.db.models import CompanyContent
from dialog_lib.embeddings.generate import generate_embedding

from langchain_community.document_loaders import WebBaseLoader


def load_webpage(url, dbsession, embeddings_model_instance, company_id=None):
    loader = WebBaseLoader(url)
    contents = loader.load()

    for url_content in contents:
        company_content = CompanyContent(
            link=url,
            category="web",
            subcategory="website-content",
            question=url_content.metadata["title"],
            content=url_content.page_content,
            dataset=company_id,
            embedding=generate_embedding(url_content.page_content, embeddings_model_instance)
        )
        dbsession.add(company_content)
        dbsession.flush()

    return company_content


class FakeEmbeddingModel:
    def embed_documents(self, contents, token_length=1536):
        return_value = []
        for content in contents:
            return_value.append([0] * token_length)

        return return_value

    def embed_query(self, content, token_length=1536):
        return [0] * token_length


from .abstract import AbstractDialog
from langchain_anthropic import ChatAnthropic


class DialogAnthropic(AbstractDialog):
    def __init__(self, *args, **kwargs):
        model = kwargs.pop("model", "claude-3-opus-20240229")
        temperature = kwargs.pop("temperature", 0)
        kwargs["model_class"] = ChatAnthropic(
            model=model,
            temperature=temperature,
            anthropic_api_key=kwargs.get("llm_api_key"),
            max_tokens=1536
        )
        super().__init__(*args, **kwargs)

    def postprocess(self, output):
        return output.get("text")


import uuid

from .models import Chat


def create_chat_session(identifier=None, dbsession=None, model=Chat):
    if identifier is None:
        identifier = uuid.uuid4().hex

    chat = dbsession.query(model).filter_by(session_id=identifier).first()
    if not chat:
        chat = model(session_id=identifier)
        dbsession.add(chat)
        dbsession.commit()

    return {"chat_id": chat.session_id}
